*! version 1.1.0 March 23, 2006 @ 11:18:59
*! Perform an optimal matching with Needleman-Wunsch-Algorithm
*! Thanks to Wikipedia.de for nice description of NW
* Source: http://de.wikipedia.org/wiki/Needleman-Wunsch-Algorithmus

program sqom
version 9.1
	syntax [if] [in] [, Name(string) INDELcost(real 1) SUBcost(string) STandard(string) ///
	  refseq(string) i(varname) t(varname) sq(varname)  k(int 0)]

	// SQ-Data
	// -------

	capture _sqset `sq' `i' `t'
	if _rc {
		di as error "data not declared as SQ-data; use -sqset-"
		exit 9
	}

	// Declarations
	// ------------

	tempvar gap epiid epicount SQom SQid cutter
	tempfile noshrink cluster results

	quietly {

		// Error Checks
		// ------------

		if "`name'" == "" capture drop _SQomcluster*
		else {
			gettoken  clustervarname : name, parse(",")
			if strpos("`name'","replace") > 0 capture drop `clustervarname'_id `clustervarname'_ord `clustervarname'_hgt
			confirm new variable `SQomcluster'_id `SQomcluster'_ord `SQomcluster'_hgt
		}
		local SQomcluster = cond("`name'"=="","_SQomcluster","`clustervarname'")

		local check = real("`subcost'")
		if `check' <= 0   {
			noi di as error "subcost(<=0) invalid"
			exit 198
		}

		// Shrink the Data
		// ---------------
		
		preserve
		marksample touse
		keep if `touse '
		keep `_dta[SQis]' `_dta[SQiis]' `_dta[SQtis]'
		
		// Idenify Sequences with Gaps -> left censor is a gap!
		// ---------------------------
		// (I delete them later on)
		
		by `_dta[SQiis]' (`_dta[SQtis]'), sort: ///
		  gen `epiid' = sum(`_dta[SQis]' != `_dta[SQis]'[_n-1])
		by `_dta[SQiis]' (`_dta[SQtis]'): ///
		  gen `epicount' = `epiid'[_N]
		gen `gap' = 1 if `_dta[SQis]'>=. & `epiid' < `epicount' // & `epiid'>1  -> left censor is a gap!
		by `_dta[SQiis]' (`_dta[SQtis]'): replace `gap' = sum(`gap')
		by `_dta[SQiis]' (`_dta[SQtis]'): replace `gap' = `gap'[_N]>0
		drop `epiid' `epicount'
		

		// Prepare the cut-Standardisation
		// -------------------------------

		capture confirm integer number `standard'
		if !_rc {
			drop if `_dta[SQtis]' > `standard'
			local standard none
		}
		else if "`standard'" == "cut" {
			drop if `_dta[SQis]' >= . & !`gap' 
			by `_dta[SQiis]' (`_dta[SQtis]'), sort: gen `cutter' = _N if !`gap'
			sum `cutter', meanonly
			drop if `_dta[SQtis]' > r(min)
			local standard none
		}
		else if "`standard'" == "" {
			local standard longest
		}
		else if "`standard'" != "none" {
			noi di as error "standard() invalid"
			exit 198
		}
		
		// Construct the matrix for the substitution costs
		// -----------------------------------------------
		
		local mod 0
		if "`subcost'" == "" local subcost = `indelcost'*2
		capture confirm number `subcost'
		if _rc {
			if "`subcost'" == "rawdistance" {
				capture confirm numeric variable `_dta[SQis]'
				if _rc {
					noi di as error "subcost(rawdistance) not allowed with string variables"
					exit 189
				}
				local subcost -1
			}
			else {
				capture matrix SQsubcost = `subcost'
				if _rc {
				noi di as error "subcost() invalid: specify number, -rawdistance- or a matrix-name"
					exit 198
				}
				levelsof `_dta[SQis]', local(Element)
				matrix rownames SQsubcost = `Element'
				matrix colnames SQsubcost = `Element'
				local subcost 0
			}
			
			// Initialize Hash Table
			// ---------------------

			capture confirm string variable `_dta[SQis]'
			if _rc local istring 0 
			else local istring 1
			
			.h = .hash.new
			local nofElements: word count `Element'
			local mod = cond(`nofElements'<100,197,6709) // mod set to prime numbers (!)
			.h.initialize `mod'
			local i 1
			foreach element of local Element {
				if !`istring' {
					local p = mod(ceil(`element'),`mod')
					local coord = `i++'
				}
				.h.setarray `p' `element' `coord'
			}
		}
		

		// Reshape Wide
		// ------------
		
		reshape wide `_dta[SQis]', i(`_dta[SQiis]') j(`_dta[SQtis]') 
		unab varlist: `_dta[SQis]'*
		
		// Shrink the data even more
		// ------------------------

		// Drop Sequences with gaps
		capture assert `gap' == 0
		if _rc {
			count if `gap' > 0
			noi di as text "Note: " as res r(N) as text " sequences with gaps have been removed from the analysis"
			drop if `gap' > 0
		}

		// Store a copy 
		by `varlist', sort: gen `SQid' = 1 if _n==1
		replace `SQid' = sum(`SQid')
		sort `SQid'
		save `noshrink'

		// Keep only one Sequence of each type
		by `SQid': keep if _n==1
		sort `_dta[SQiis]'

		// Call the Mata-Function from lsq.lib (Source-Code: lsq.mata)
		// ----------------------------------------------------------

		noi di as text "Perform " as res _N*(_N-1)/2 as text " Comparison with Needleman-Wunsch Algorithm"
		mata: sqom("`varlist'",`indelcost',"`standard'",`k',`subcost',`mod')
		noi di as text "Distance matrix saved as " as res "SQdist" 
		
		// Perform Cluster-Analysis and Merge Results to Data
		// --------------------------------------------------

		clustermat wardslinkage SQdist, name(`SQomcluster') clear

		gen `SQid' = `SQomcluster'_id 
		sort `SQid'
		save `cluster'
		
		use `noshrink', replace
		merge `SQid' using `cluster'
		keep `_dta[SQiis]' `SQomcluster'*
		sort `_dta[SQiis]'

		save `results'
		
		restore 
		sort `_dta[SQiis]'
		merge `_dta[SQiis]' using `results'
		drop _merge

		noi di as text "Results of cluster analysis saved as " as res "`SQomcluster'_*"
		char _dta[SQcluster_ord] `SQomcluster'_ord 
	}
	end
		
// Not used, actually. I is slower than subcost. 
// (Subprogramm zum Auslesen der Kosten aus der Hash-Tabelle. Wird von subcosthash() aufgerufen)
// `1' ist Key1, `2' is Key2 und `3` ist der Modulo-Wert
	
program SQGetCost
	.adress1 = .h.search `=mod(`1',`3')' `1'
	.adress2 = .h.search `=mod(`2',`3')' `2'
	scalar SQsubcostvalue = SQsubcost[`.adress1',`.adress2']
end
	
